#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 10 15:52:23 2022

@author: nana
"""
import os, time, h5py
import numpy as np
from .image_processing import*
np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)

def bin_2_rtdc(bin_path):
    t1 = time.time()
    ###############Parameters##########################
    filter_len = True
    len_min = 10
    len_max = 300
    
    filter_area = False
    area_min = 10
    area_max = 50
    
    filter_n = True
    nr_contours = 100
    
    pixel_size = 0.8
    noise_level = 8
    bg_intensity = 16381
    ###############Parameters##########################
    
    path_target = os.path.dirname(bin_path) + os.sep + os.path.basename(bin_path).split(".bin")[0] + ".rtdc" #filename of resulting file
    #check if the file that should be written may already exists (e.g. after runnig script twice)
    if os.path.isfile(path_target):
        print("Following file already exists and will be overwritten: "+path_target)
        #delete the file
        os.remove(path_target)
    
    
    #load binary
    binary = np.fromfile(bin_path, dtype='>H')
    n,w,h = binary[1],binary[3],binary[5]
    images = binary[6:].reshape(n,h,w)
    
    images = images.astype(np.float32) #for conversion to unit8, first make it float (float32 is sufficient)
    factor = 128/bg_intensity
    images = np.multiply(images, factor)
    images = images.astype(np.uint8)

    #Cell can be darker and brighter than background->Subtract background and take absolute
    images_abs = images.astype(np.int32)-128 #after that, the background is approx 0
    images_abs = abs(images_abs).astype(np.uint8) #absolute. Neccessary for contour finding
    

    #Initialize lists for all properties
    index_orig,images_save,masks,contours,\
    pos_x,pos_y,size_x,size_y,\
    bright_avg,bright_sd,\
    area_um,area_orig,area_hull,\
    area_ratio,circularity,inert_ratio_raw = \
    [],[],[],[],[],[],[],[],[],[],[],[],[],[],[],[]
    
    t1 = time.time()
    #get all images located in one particular folder
    for img_index in range(len(images)):
        #load image
        image = images[img_index]
        ####
        #image, factor = uint16_2_unit8(image) 
        image_abs = images_abs[img_index]
                
        #get list of conoturs and masks in image using the superposition
        contours_,masks_ = get_masks_ipac(image_abs, noise_level,filter_len, len_min, len_max, 
                                             filter_area, area_min, area_max, filter_n, nr_contours)
        del image_abs
        
        for contour,mask in zip(contours_,masks_):
            output = get_boundingbox_features(image,contour,pixel_size)
    
            if type(contour)!=np.ndarray or np.isnan(output[0]) or np.isnan(output[1]):
                index_orig.append(img_index)
                images_save.append(np.zeros(shape=image.shape,dtype=np.uint8))
                masks.append(np.zeros(shape=image.shape))
                contours.append(np.nan)
                pos_x.append(np.nan)
                pos_y.append(np.nan)
                size_x.append(np.nan)
                size_y.append(np.nan)
    
                bright_avg.append(np.nan)
                bright_sd.append(np.nan)
                area_orig.append(np.nan)
                area_hull.append(np.nan)
                area_um.append(np.nan)
                area_ratio.append(np.nan)
                circularity.append(np.nan)
                inert_ratio_raw.append(np.nan)
    
            else:
                index_orig.append(img_index)
                images_save.append(image)
                masks.append(mask)
                contours.append(contour)
                pos_x.append(output[0])
                pos_y.append(output[1])
                size_x.append(output[2])
                size_y.append(output[3])
                
                output = get_brightness(image,mask)
                bright_avg.append(output["bright_avg"])
                bright_sd.append(output["bright_sd"])
                
                output = get_contourfeatures(contour,pixel_size)
                area_orig.append(output["area_orig"])
                area_hull.append(output["area_hull"])
                area_um.append(output["area_um"])
                area_ratio.append(output["area_ratio"])
                circularity.append(output["circularity"])
                inert_ratio_raw.append(output["inert_ratio_raw"])
    
    t2 = time.time()
    dt = t2-t1
    print("Required time to compute features: " +str(np.round(dt,2) )+"s ("+str(np.round(dt/len(images_save)*1000,2) )+"ms per cell)")
    
    
    #remove events where no contours were found (pos_x and pos_y is nan)
    ind_nan = np.isnan(pos_x)
    ind_nan = np.where(ind_nan==False)[0]
    if len(ind_nan)>0:
        index_orig = list(np.array(index_orig)[ind_nan])
        pos_x = list(np.array(pos_x)[ind_nan])
        pos_y = list(np.array(pos_y)[ind_nan])
        size_x = list(np.array(size_x)[ind_nan])
        size_y = list(np.array(size_y)[ind_nan])
    
# =============================================================================
#         images_save = list(np.array(images_save,dtype=np.uint8)[ind_nan])
#         masks = list(np.array(masks,dtype=np.bool_)[ind_nan])   
# =============================================================================
        images_save = [images_save[index] for index in ind_nan]
        masks = [masks[index] for index in ind_nan]
        
        contours = list(np.array(contours)[ind_nan])
        bright_avg = list(np.array(bright_avg)[ind_nan])
        bright_sd = list(np.array(bright_sd)[ind_nan])
        area_orig = list(np.array(area_orig)[ind_nan])
        area_hull = list(np.array(area_hull)[ind_nan])
        area_um = list(np.array(area_um)[ind_nan])
        area_ratio = list(np.array(area_ratio)[ind_nan])
        circularity = list(np.array(circularity)[ind_nan])
        inert_ratio_raw = list(np.array(inert_ratio_raw)[ind_nan])
        
        #Save images and corresponding pos_x and pos_y to an hdf5 file for AIDeveloper
        images_save = np.array(images_save)
        masks = np.array(masks)
        
        maxshape_img = (None, images_save.shape[1], images_save.shape[2])
        maxshape_mask = (None, masks.shape[1], masks.shape[2])
        
        #Create rtdc_dataset; valid feature names can be found via dclab.dfn.feature_names
        hdf = h5py.File(path_target,'a')
        dset = hdf.create_dataset("events/image", data=images_save, dtype=np.uint8,maxshape=maxshape_img,fletcher32=True,chunks=True)
        dset.attrs.create('CLASS', np.string_('IMAGE'))
        dset.attrs.create('IMAGE_VERSION', np.string_('1.2'))
        dset.attrs.create('IMAGE_SUBCLASS', np.string_('IMAGE_GRAYSCALE'))
    
        dset = hdf.create_dataset("events/mask", data=masks, dtype=np.uint8,maxshape=maxshape_mask,fletcher32=True,chunks=True)
        dset.attrs.create('CLASS', np.string_('IMAGE'))
        dset.attrs.create('IMAGE_VERSION', np.string_('1.2'))
        dset.attrs.create('IMAGE_SUBCLASS', np.string_('IMAGE_GRAYSCALE'))
        
        hdf.create_dataset("events/index_online", data=index_orig,dtype=np.int32)
        hdf.create_dataset("events/pos_x", data=pos_x, dtype=np.int32)
        hdf.create_dataset("events/pos_y", data=pos_y, dtype=np.int32)
        hdf.create_dataset("events/size_x", data=size_x, dtype=np.int32)
        hdf.create_dataset("events/size_y", data=size_y, dtype=np.int32)
        
        hdf.create_dataset("events/bright_avg", data=bright_avg, dtype=np.float32)
        hdf.create_dataset("events/bright_sd", data=bright_sd, dtype=np.float32)
    
        hdf.create_dataset("events/circ", data=circularity, dtype=np.float32)
        hdf.create_dataset("events/inert_ratio_raw", data=inert_ratio_raw, dtype=np.float32)
        hdf.create_dataset("events/area_ratio", data=area_ratio, dtype=np.float32)
        hdf.create_dataset("events/area_msd", data=area_orig, dtype=np.float32)
        hdf.create_dataset("events/area_cvx", data=area_hull, dtype=np.float32)       
        hdf.create_dataset("events/area_um", data=area_um, dtype=np.float32)       
    
        #Adjust metadata:
# =============================================================================
#                 #"experiment:event count" = Nr. of images
#                 hdf.attrs["experiment:run index"] = m_number
#                 m_number += 1 #increase measurement number 
# =============================================================================
        hdf.attrs["experiment:event count"] = images_save.shape[0]
        #hdf.attrs["experiment:sample"] = condition #Blood draw date
        hdf.attrs["imaging:pixel size"] = pixel_size
        hdf.attrs["experiment:date"] = time.strftime("%Y-%m-%d")
        hdf.attrs["experiment:time"] = time.strftime("%H:%M:%S")

        hdf.attrs["imaging:roi size x"] = images_save.shape[2]
        hdf.attrs["imaging:roi size y"] = images_save.shape[1]
        hdf.attrs["online_contour:bin kernel"] = 2*int(3)+1
        hdf.attrs["online_contour:bin threshold"] = noise_level

        hdf.close()
    
    t2 = time.time()
    print("bin. to rtdc.:",np.round((t2-t1),4))
    
    return path_target


# =============================================================================
# def bin_2_rtdc(bin_path):
#     t3 = time.time()
#     ###############Parameters##########################
#     filter_len = True
#     len_min = 10
#     len_max = 300
#     
#     filter_area = False
#     area_min = 10
#     area_max = 50
#     
#     filter_n = True
#     nr_contours = 100
#     
#     pixel_size = 0.8
#     noise_level = 8
#     bg_intensity = 16381
#     ###############Parameters##########################
#     
#     path_target = os.path.dirname(bin_path) + os.sep + os.path.basename(bin_path).split(".bin")[0] + ".rtdc" #filename of resulting file
#     #check if the file that should be written may already exists (e.g. after runnig script twice)
#     if os.path.isfile(path_target):
#         print("Following file already exists and will be overwritten: "+path_target)
#         #delete the file
#         os.remove(path_target)
#     
#     
#     #load binary
#     binary = np.fromfile(bin_path, dtype='>H')
#     n,w,h = binary[1],binary[3],binary[5]
#     images = binary[6:].reshape(n,h,w)
#     
#     images = images.astype(np.float32) #for conversion to unit8, first make it float (float32 is sufficient)
#     factor = 128/bg_intensity
#     images = np.multiply(images, factor)
#     images = images.astype(np.uint8)
#     
#     #Cell can be darker and brighter than background->Subtract background and take absolute
#     images_abs = images.astype(np.int32)-128 #after that, the background is approx 0
#     images_abs = abs(images_abs).astype(np.uint8) #absolute. Neccessary for contour finding
#     
#     
#     #Initialize lists for all properties
#     index_orig,images_save,masks,contours,\
#     pos_x,pos_y,size_x,size_y,\
#     bright_avg,bright_sd,\
#     area_um,area_orig,area_hull,\
#     area_ratio,circularity,inert_ratio_raw = \
#     [],[],[],[],[],[],[],[],[],[],[],[],[],[],[],[]
#     
#     t4 = time.time()
#     print("Load binary takes:",np.round((t4-t3),2),"s")
#     
#     t1 = time.time()
#     #get all images located in one particular folder
#     for img_index in range(len(images)):
#         #load image
#         image = images[img_index]
#         ####
#         #image, factor = uint16_2_unit8(image) 
#         image_abs = images_abs[img_index]
#                 
#         #get list of conoturs and masks in image using the superposition
#         contours_,masks_ = get_masks_ipac(image_abs, noise_level,filter_len, len_min, len_max, 
#                                              filter_area, area_min, area_max, filter_n, nr_contours)
#         del image_abs
#         
#         for contour,mask in zip(contours_,masks_):
#             output = get_boundingbox_features(image,contour,pixel_size)
#     
#             if type(contour)!=np.ndarray or np.isnan(output[0]) or np.isnan(output[1]):
#                 index_orig.append(img_index)
#                 images_save.append(np.zeros(shape=image.shape))
#                 masks.append(np.zeros(shape=image.shape))
#                 contours.append(np.nan)
#                 pos_x.append(np.nan)
#                 pos_y.append(np.nan)
#                 size_x.append(np.nan)
#                 size_y.append(np.nan)
#     
#                 bright_avg.append(np.nan)
#                 bright_sd.append(np.nan)
#                 area_orig.append(np.nan)
#                 area_hull.append(np.nan)
#                 area_um.append(np.nan)
#                 area_ratio.append(np.nan)
#                 circularity.append(np.nan)
#                 inert_ratio_raw.append(np.nan)
#     
#             else:
#                 index_orig.append(img_index)
#                 images_save.append(image)
#                 masks.append(mask)
#                 contours.append(contour)
#                 pos_x.append(output[0])
#                 pos_y.append(output[1])
#                 size_x.append(output[2])
#                 size_y.append(output[3])
#                 
#                 output = get_brightness(image,mask)
#                 bright_avg.append(output["bright_avg"])
#                 bright_sd.append(output["bright_sd"])
#                 
#                 output = get_contourfeatures(contour,pixel_size)
#                 area_orig.append(output["area_orig"])
#                 area_hull.append(output["area_hull"])
#                 area_um.append(output["area_um"])
#                 area_ratio.append(output["area_ratio"])
#                 circularity.append(output["circularity"])
#                 inert_ratio_raw.append(output["inert_ratio_raw"])
#     
#     t2 = time.time()
#     dt = t2-t1
#     print("Required time to compute features: " +str(np.round(dt,2) )+"s ("+str(np.round(dt/len(images_save)*1000,2) )+"ms per cell)")
#     
#     
#     
#     #remove events where no contours were found (pos_x and pos_y is nan)
#     ind_nan = np.isnan(pos_x)
#     ind_nan = np.where(ind_nan==False)[0]
#     if len(ind_nan)>0:
#         index_orig = list(np.array(index_orig)[ind_nan])
#         pos_x = list(np.array(pos_x)[ind_nan].astype(dtype=np.int32))
#         pos_y = list(np.array(pos_y)[ind_nan].astype(dtype=np.int32))
#         size_x = list(np.array(size_x)[ind_nan].astype(dtype=np.int32))
#         size_y = list(np.array(size_y)[ind_nan].astype(dtype=np.int32))
# 
#         images_save = list(np.array(images_save,dtype=np.uint8)[ind_nan])
#         masks = list(np.array(masks,dtype=np.bool_)[ind_nan])    
#         
#         #contours = list(np.array(contours)[ind_nan])
#         bright_avg = list(np.array(bright_avg)[ind_nan].astype(dtype=np.float32))
#         bright_sd = list(np.array(bright_sd)[ind_nan].astype(dtype=np.float32))
#         area_orig = list(np.array(area_orig)[ind_nan].astype(dtype=np.float32))
#         area_hull = list(np.array(area_hull)[ind_nan].astype(dtype=np.float32))
#         area_um = list(np.array(area_um)[ind_nan].astype(dtype=np.float32))
#         area_ratio = list(np.array(area_ratio)[ind_nan].astype(dtype=np.float32))
#         circularity = list(np.array(circularity)[ind_nan].astype(dtype=np.float32))
#         inert_ratio_raw = list(np.array(inert_ratio_raw)[ind_nan].astype(dtype=np.float32))
#         
#         
#         
#         #Save images and corresponding pos_x and pos_y to an hdf5 file for AIDeveloper
#         images_save = np.r_[images_save]
#         masks = np.r_[masks]
#         
#         maxshape_img = (None, images_save.shape[1], images_save.shape[2])
#         maxshape_mask = (None, masks.shape[1], masks.shape[2])
#         
#         #create a dictionary
#         dic = {}
#         dic["image"] = images_save
#         dic["mask"] = masks
#         dic["index_online"] = index_orig
#         dic["pos_x"] = pos_x
#         dic["pos_y"] = pos_y
#         dic["size_x"] = size_x
#         dic["size_y"] = size_y
#         dic["bright_avg"] = bright_avg
#         dic["bright_sd"] = bright_sd
#         dic["circ"] = circularity
#         dic["inert_ratio_raw"] = inert_ratio_raw
#         dic["area_ratio"] = area_ratio
#         dic["area_msd"] = area_orig
#         dic["area_cvx"] = area_hull
#         dic["area_um"] = area_um
#         dic["imaging:pixel size"] = pixel_size
#         
#         
#     t4 = time.time()
#     #print("write .rtdc.:",np.round((t4-t2),2))
#     print("remove events with no contours:",np.round((t4-t2),2))
#     print("bin. to rtdc.:",np.round((t4-t3),2))  
#     
#     return dic
# =============================================================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools

setuptools.setup(
      packages=setuptools.find_namespace_packages(
                     include=["iacs_ipac_reader", "iacs_ipac_reader.*"], ),
                 include_package_data=True
                 )


import numpy as np
import pandas as pd
import cv2
#from scipy import ndimage
#import aid_cv2_dnn

#Run inference (using OpenCV's dnn module) on images using your model trained 
#in AIDeveloper.
#Script contains functions: 
#- to pre-preprocess images according to the input criteria of a loaded model
#- forward images through a model and get output (predictions)


def image_adjust_channels(images,target_channels=1):
    """
    Check the number of channels of images.
    Transform images (if needed) to get to the desired number of channels
    
    Parameters
    ----------
    images: numpy array of dimension (nr.images,height,width) for grayscale,
    or of dimension (nr.images,height,width,channels) for RGB images
    target_channels: int
        target number of channels
        can be one of the following:
        
        - 1: target is a grayscale image. In case RGB images are 
        provided, the luminosity formula is used to convert of RGB to 
        grayscale
        - 3: target is an RGB image. In case grayscale images are provided,
        the information of each image is copied to all three channels to 
        convert grayscale to RGB"
    
    Returns
    ----------
    images: numpy array
        images with adjusted number of channels
    """

    #images.shape is (N,H,W) for grayscale, or (N,H,W,C) for RGB images
    #(N,H,W,C) means (nr.images,height,width,channels)

    #Make sure either (N,H,W), or (N,H,W,C) is provided
    assert len(images.shape)==4 or len(images.shape)==3, "Shape of 'images' \
    is not supported: " +str(images.shape) 

    if len(images.shape)==4:#Provided images are RGB
        #Mare sure there are 3 channels (RGB)
        assert images.shape[-1]==3, "Images have "+str(images.shape[-1])+" channels. This is (currently) not supported!"

        if target_channels==1:#User wants Grayscale -> use the luminosity formula
            images = (0.21 * images[:,:,:,:1]) + (0.72 * images[:,:,:,1:2]) + (0.07 * images[:,:,:,-1:])
            images = images[:,:,:,0] 
            images = images.astype(np.uint8)           
            print("Used luminosity formula to convert RGB to Grayscale")
            
    if len(images.shape)==3:#Provided images are Grayscale
        if target_channels==3:#User wants RGB -> copy the information to all 3 channels
            images = np.stack((images,)*3, axis=-1)
            print("Copied information to all three channels to convert Grayscale to RGB")
    return images


def image_crop_pad_np(images,pos_x,pos_y,final_h,final_w,padding_mode='constant'):
    """
    Deprecated: Please use 'image_crop_pad_cv' instead, which uses OpenCV instead of numpy.
    
    Function takes a list images (list of numpy arrays) an resizes them to 
    equal size by center cropping and/or padding.
    Parameters
    ----------
    images: list of images of arbitrary shape
    (nr.images,height,width,channels) 
        can be a single image or multiple images
    pos_x: float or ndarray of length N
        The x coordinate(s) of the centroid of the event(s) [um]
    pos_y: float or ndarray of length N
        The y coordinate(s) of the centroid of the event(s) [um]
    final_h: int
        target image height [pixels]
    final_w: int
        target image width [pixels]
    padding_mode: str
        Perform the following padding operation if the cell is too far at the 
        border of the image such that the  desired image size cannot be 
        obtained without going beyond the order of the image:
        
        - Delete: Return empty array (all zero) if the cell is too far at border (delete image)
        
        #the following text is copied from 
        https://docs.scipy.org/doc/numpy/reference/generated/numpy.pad.html</p></body></html>"
        
        - constant (default): Pads with a constant value.
        - edge: Pads with the edge values of array.
        - linear_ramp: Pads with the linear ramp between end_value and the array edge value.
        - maximum: Pads with the maximum value of all or part of the vector along each axis.
        - mean: Pads with the mean value of all or part of the vector along each axis.
        - median: Pads with the median value of all or part of the vector along each axis.
        - minimum: Pads with the minimum value of all or part of the vector along each axis.
        - reflect: Pads with the reflection of the vector mirrored on the first and last values of the vector along each axis.
        - symmetric: Pads with the reflection of the vector mirrored along the edge of the array.
        - wrap: Pads with the wrap of the vector along the axis. The first values are used to pad the end and the end values are used to pad the beginning.
    
    Returns
    ----------
    images: list of images. Each image is a numpy array of shape 
        (final_h,final_w,channels) 
    """
    print("Deprecated: Please use 'image_crop_pad_cv' instead, which uses OpenCV instead of numpy.")
    
    for i in range(len(images)):
        image = images[i]
    
        #Compute the edge-coordinates that define the cropped image
        y1 = np.around(pos_y[i]-final_h/2.0)              
        x1 = np.around(pos_x[i]-final_w/2.0) 
        y2 = y1+final_h               
        x2 = x1+final_w

        #Are these coordinates within the oringinal image?
        #If not, the image needs padding
        pad_top,pad_bottom,pad_left,pad_right = 0,0,0,0

        if y1<0:#Padding is required on top of image
            pad_top = int(abs(y1))
            y1 = 0 #set y1 to zero and pad pixels after cropping
            
        if y2>image.shape[0]:#Padding is required on bottom of image
            pad_bottom = int(y2-image.shape[0])
            y2 = image.shape[0]
        
        if x1<0:#Padding is required on left of image
            pad_left = int(abs(x1))
            x1 = 0
        
        if x2>image.shape[1]:#Padding is required on right of image
            pad_right = int(x2-image.shape[1])
            x2 = image.shape[1]
        
        #Get cropped image
        temp = image[int(y1):int(y2),int(x1):int(x2)]

        if pad_top+pad_bottom+pad_left+pad_right>0:
            if padding_mode=="Delete":
                temp = np.zeros_like(temp)
            else:
                #Perform all padding operations in one go
                temp = np.pad(temp,pad_width=( (pad_top, pad_bottom),(pad_left, pad_right) ),mode=padding_mode)

        images[i] = temp
            
    return images

def image_crop_pad_cv2(images,pos_x,pos_y,pix,final_h,final_w,padding_mode="cv2.BORDER_CONSTANT"):
    """
    Function takes a list images (list of numpy arrays) an resizes them to 
    equal size by center cropping and/or padding.
    Parameters
    ----------
    images: list of images of arbitrary shape
    (nr.images,height,width,channels) 
        can be a single image or multiple images
    pos_x: float or ndarray of length N
        The x coordinate(s) of the centroid of the event(s) [um]
    pos_y: float or ndarray of length N
        The y coordinate(s) of the centroid of the event(s) [um]
        
    final_h: int
        target image height [pixels]
    
    final_w: int
        target image width [pixels]
        
    padding_mode: str; OpenCV BorderType
        Perform the following padding operation if the cell is too far at the 
        border such that the  desired image size cannot be 
        obtained without going beyond the order of the image:
        
        - "Delete": Return empty array (all zero) if the cell is too far at border (delete image)
        
        #the following text is copied from 
        https://docs.opencv.org/3.4/d2/de8/group__core__array.html#ga209f2f4869e304c82d07739337eae7c5
        
        - "cv2.BORDER_CONSTANT": iiiiii|abcdefgh|iiiiiii with some specified i 
        - "cv2.BORDER_REFLECT": fedcba|abcdefgh|hgfedcb
        - "cv2.BORDER_REFLECT_101": gfedcb|abcdefgh|gfedcba
        - "cv2.BORDER_DEFAULT": same as BORDER_REFLECT_101
        - "cv2.BORDER_REPLICATE": aaaaaa|abcdefgh|hhhhhhh
        - "cv2.BORDER_WRAP": cdefgh|abcdefgh|abcdefg
    
    Returns
    ----------
    images: list of images. Each image is a numpy array of shape 
    (final_h,final_w,channels) 
    """
    #Convert position of cell from "um" to "pixel index"
    pos_x,pos_y = pos_x/pix,pos_y/pix  
    
    
    for i in range(len(images)):
        image = images[i]
    
        #Compute the edge-coordinates that define the cropped image
        y1 = np.around(pos_y[i]-final_h/2.0)              
        x1 = np.around(pos_x[i]-final_w/2.0) 
        y2 = y1+final_h               
        x2 = x1+final_w

        #Are these coordinates within the oringinal image?
        #If not, the image needs padding
        pad_top,pad_bottom,pad_left,pad_right = 0,0,0,0

        if y1<0:#Padding is required on top of image
            pad_top = int(abs(y1))
            y1 = 0 #set y1 to zero and pad pixels after cropping
            
        if y2>image.shape[0]:#Padding is required on bottom of image
            pad_bottom = int(y2-image.shape[0])
            y2 = image.shape[0]
        
        if x1<0:#Padding is required on left of image
            pad_left = int(abs(x1))
            x1 = 0
        
        if x2>image.shape[1]:#Padding is required on right of image
            pad_right = int(x2-image.shape[1])
            x2 = image.shape[1]
        
        #Crop the image
        temp = image[int(y1):int(y2),int(x1):int(x2)]

        if pad_top+pad_bottom+pad_left+pad_right>0:
            if padding_mode=="Delete":
                temp = np.zeros_like(temp)
            else:
                #Perform all padding operations in one go
                temp = cv2.copyMakeBorder(temp, pad_top, pad_bottom, pad_left, pad_right, eval(padding_mode))
        
        images[i] = temp
            
    return images


def image_zooming(images,zoom_factor,zoom_interpol_method):
    """
    Function takes a list of images (list of numpy arrays) an resizes them to 
    an equal size by scaling (interpolation).
    Parameters
    ----------
    images: list of images of arbitrary shape
    zoom_factor: float
        Factor by which the size of the images should be zoomed
    zoom_interpol_method: str; available are: (text copied from original docs: https://docs.opencv.org/2.4/modules/imgproc/doc/geometric_transformations.html#resize)
        -"cv2.INTER_NEAREST" – a nearest-neighbor interpolation
        -"cv2.INTER_LINEAR" – a bilinear interpolation (used by default)
        -"cv2.INTER_AREA" – resampling using pixel area relation. It may be a preferred method for image decimation, as it gives moire’-free results. But when the image is zoomed, it is similar to the INTER_NEAREST method.
        -"cv2.INTER_CUBIC" - a bicubic interpolation over 4x4 pixel neighborhood
        -"cv2.INTER_LANCZOS4" - a Lanczos interpolation over 8x8 pixel neighborhood
    Returns
    ----------
    list of images of arbitrary shape
    """

    #final_h = int(np.around(zoom_factor*images[0].shape[1]))
    #final_w = int(np.around(zoom_factor*images[0].shape[2]))

    for i in range(len(images)):
        #the order (width,height) in cv2.resize is not an error. OpenCV wants this order...
        #images[i] = cv2.resize(images[i], dsize=(final_w,final_h), interpolation=eval(zoom_interpol_method))
        images[i] = cv2.resize(images[i], dsize=None,fx=zoom_factor, fy=zoom_factor, interpolation=eval(zoom_interpol_method))

    return images

def image_normalization(images,normalization_method,mean_trainingdata=None,std_trainingdata=None):
    """
    Perform a normalization of the pixel values.
    Parameters
    ----------
    images: ndarray
    normalization_method: str
        Factor by which the size of the images should be zoomed
    normalization_method: str; available are: (text copied from original docs: 
        https://docs.opencv.org/2.4/modules/imgproc/doc/geometric_transformations.html#resize)
        -"None" – No normalization is applied.
        -"Div. by 255" – Each input image is divided by 255 (useful since pixel 
        values go from 0 to 255, so the result will be in range 0-1)
        -"StdScaling using mean and std of each image individually" – The mean 
        and standard deviation of each input image itself is used to scale it 
        by first subtracting the mean and then dividing by the standard deviation
        -"StdScaling using mean and std of all training data" - During model 
        training, the mean and std of the entire training set was determined. 
        This mean and standard deviation is used to normalize images by first 
        subtracting the mean and then dividing by the standard deviation    
    mean_trainingdata: float; the mean pixel value obtained from the training dataset
    std_trainingdata: float; the std of the pixel values obtained from the training dataset
    Returns
    ----------
    ndarray of images
  
    """
    if normalization_method == "StdScaling using mean and std of all training data":
        #make sure pandas series is converted to numpy array
        if type(mean_trainingdata)==pd.core.series.Series:
            mean_trainingdata=mean_trainingdata.values[0]
            std_trainingdata=std_trainingdata.values[0]
        if np.allclose(std_trainingdata,0):
            std_trainingdata = 0.0001
            print("Set the standard deviation (std_trainingdata) to 0.0001 because otherwise div. by 0 would have happend!")

    if len(images.shape)==3: #single channel Grayscale rtdc data
        #Add the "channels" dimension
        images = np.expand_dims(images,3)    
    images = images.astype(np.float32)
    
    for k in range(images.shape[0]):
        line = images[k,:,:,:]
        ###########Scaling############
        if normalization_method == "None":
            pass #dont do anything
        elif normalization_method == "Div. by 255":
            line = line/255.0
        elif normalization_method == "StdScaling using mean and std of each image individually":
            mean = np.mean(line)
            std = np.std(line)
            if np.allclose(std,0):
                std = 0.0001
                print("Set the standard deviation to 0.0001 because otherwise div. by 0 would have happend!")
            line = (line-mean)/std
        elif normalization_method == "StdScaling using mean and std of all training data":
            line = (line-mean_trainingdata)/std_trainingdata
            
        #Under NO circumstances, training data should contain nan values
        ind = np.isnan(line)
        line[ind] = np.random.random() #replace nan with random values. This is better than nan, since .fit will collapse and never get back
        images[k,:,:,:] = line   
    return images        


def image_preprocessing(images,pos_x,pos_y,pix=0.34,target_imsize=32,
                        target_channels=1,zoom_factor=1,
                        zoom_interpol_method="cv2.INTER_NEAREST",
                        padding_mode="cv2.BORDER_CONSTANT",
                        normalization_method="Div. by 255",
                        mean_trainingdata=None,std_trainingdata=None):
    """
    Wrapper function which performs all image preprocessing steps, required
    for preparing raw rtdc images to be forwarded through a neural net,
    
    Parameters
    ----------
    images: numpy array of shape (nr.images,height,width,channels) 
        can be a single image or multiple images
    pos_x: float or ndarray of length N
        The x coordinate(s) of the centroid of the event(s) [um]
    pos_y: float or ndarray of length N
        The y coordinate(s) of the centroid of the event(s) [um]
    pix: float
        Resolution [µm/pix]
    target_imsize: int
        target image size (in pixels)
        currently, only squared images are supported. Hence, width=height
    target_channels: int
        Indicates the number of channels of the images
        can be one of the following:
        - 1: model expects grayscale images 
        - 3: model expects RGB (color) images 
    zoom_factor: float
        Factor by which the size of the images should be zoomed
    zoom_interpol_method: str; OpenCV interpolation flag
        can be one of the following methods:
        - "cv2.INTER_NEAREST": a nearest-neighbor interpolation 
        - "cv2.INTER_LINEAR": a bilinear interpolation (used by default) 
        - "cv2.INTER_AREA": resampling using pixel area relation. It may be a 
        preferred method for image decimation, as it gives moire’-free results. 
        But when the image is zoomed, it is similar to the INTER_NEAREST method. 
        - "cv2.INTER_CUBIC": a bicubic interpolation over 4×4 pixel neighborhood 
        - "cv2.INTER_LANCZOS4": a Lanczos interpolation over 8×8 pixel neighborhood        
    padding_mode: str; OpenCV BorderType
        Perform the following padding operation if the cell is too far at the 
        border of the image such that the  desired image size cannot be 
        obtained without going beyond the order of the image:
        
        - "Delete": Return empty array (all zero) if the cell is too far at border (delete image)
        
        #the following text is copied from 
        https://docs.opencv.org/3.4/d2/de8/group__core__array.html#ga209f2f4869e304c82d07739337eae7c5
        
        - "cv2.BORDER_CONSTANT": iiiiii|abcdefgh|iiiiiii with some specified i 
        - "cv2.BORDER_REFLECT": fedcba|abcdefgh|hgfedcb
        - "cv2.BORDER_REFLECT_101": gfedcb|abcdefgh|gfedcba
        - "cv2.BORDER_DEFAULT": same as BORDER_REFLECT_101
        - "cv2.BORDER_REPLICATE": aaaaaa|abcdefgh|hhhhhhh
        - "cv2.BORDER_WRAP": cdefgh|abcdefgh|abcdefg
    
    normalization_method: str
        Define a method to normalize the pixel-values of the images.
        can be one of the following methods:
        - 'None': No normalization is applied
        - 'Div. by 255': each pixel value is divided by 255 (default)
        - 'StdScaling using mean and std of each image individually': The mean and 
        standard deviation of each input image itself is used to scale it by 
        first subtracting the mean and then dividing by the standard deviation.
        -StdScaling using mean and std of all training data: A mean and std. 
        value was obtained while fitting the neural net by averaging the entire 
        training dataset. These fixed values are used to scale images during 
        training by first subtracting the mean and then dividing by the 
        standard deviation.
    mean_trainingdata: float; the mean pixel value obtained from the training dataset
    std_trainingdata: float; the std of the pixel values obtained from the training dataset
    """

    #Adjust number of channels
    images = image_adjust_channels(images,target_channels)
    #Convert image array to list clea
    images = list(images)

    #Apply zooming operation if required
    if zoom_factor!=1:
        images = image_zooming(images,zoom_factor,zoom_interpol_method)
        #Adjust pos_x and pos_y accordingly
        pos_x,pos_y = zoom_factor*pos_x,zoom_factor*pos_y

    #Cropping and padding operation to obtain images of desired size
    images = image_crop_pad_cv2(images=images,pos_x=pos_x,pos_y=pos_y,pix=pix,final_h=target_imsize,final_w=target_imsize,padding_mode=padding_mode)

    #Convert to unit8 arrays
    images = np.array((images), dtype="uint8")

    if target_channels==1:#User wants Grayscale -> add empty channel dimension (required for inference in OpenCV dnn)
        images = np.expand_dims(images,3)

    #Normalize images
    images = image_normalization(images,normalization_method="Div. by 255",mean_trainingdata=None,std_trainingdata=None)

    return images

def load_model_meta(meta_path):
    """
    Extract meta information from a meta file that was created during training 
    in AID. Function returns all information how images need ot be preprocessed 
    before passing them throught the neural net. 
    Parameters
    ----------    
    meta_path: str; path to a meta file (generated by AID when during model training)
    Returns
    ----------    
    pd.DataFrame ; A DataFrame with the following keys:
        target_imsize: input image size required by the neural net
        target_channels: number of image channels required by the neural net
        normalization_method: the method to normalize the pixel-values of the images
        mean_trainingdata: the mean pixel value obtained from the training dataset
        std_trainingdata: the std of the pixel values obtained from the training dataset
        zoom_factor: factor by which the size of the images should be zoomed
        zoom_interpol_method: OpenCV interpolation flag
        padding_mode: OpenCV borderType flag
    """
    xlsx = pd.ExcelFile(meta_path)
    
    #The zooming factor is saved in the UsedData sheet
    meta = pd.read_excel(xlsx,sheet_name="UsedData")
    zoom_factor = meta["zoom_factor"].iloc[0]#should images be zoomed before forwarding through neural net?

    meta = pd.read_excel(xlsx,sheet_name="Parameters")

    try:
        model_type = meta["Chosen Model"].iloc[0]#input dimensions of the model
    except:
        model_type = "Unknown"
    
    try:
        target_imsize = meta["Input image crop"].iloc[0]#input dimensions of the model
    except:
        target_imsize = meta["Input image size"].iloc[0]#input dimensions of the model

    normalization_method = meta["Normalization"].iloc[0]#normalization method
    if normalization_method == "StdScaling using mean and std of all training data":                                
        mean_trainingdata = meta["Mean of training data used for scaling"]
        std_trainingdata = meta["Std of training data used for scaling"]
    else:
        mean_trainingdata = None
        std_trainingdata = None
  
    #Following parameters may not exist in meta files of older AID versions. Hence try/except

    #Color mode: grayscale or RGB?
    try:
        target_channels = meta["Color Mode"].iloc[0]
    except:
        target_channels = "grayscale"
    if target_channels.lower() =="grayscale":
        target_channels = 1
    elif target_channels.lower() =="rgb":
        target_channels = 3

    #The order for the zooming operation
    try:
        zoom_interpol_method = meta["Zoom order"].iloc[0]
    except:
        zoom_interpol_method = "cv2.INTER_NEAREST"
    #Translate zoom_interpol_method to OpenCV argument
    if "cv2." not in str(zoom_interpol_method):
        zoom_interpol_method = zoom_arguments_scipy2cv(zoom_factor,zoom_interpol_method)

    #Padding mode
    try:
        padding_mode = meta["paddingMode"].iloc[0]
    except:
        padding_mode = "constant"#cv2.BORDER_CONSTANT
    #translate padding_mode to OpenCV argument
    if "cv2." not in padding_mode:
        padding_mode = pad_arguments_np2cv(padding_mode)

    #Write information in one DataFrame
    img_processing_settings = pd.DataFrame()
    img_processing_settings["model_type"]=model_type,
    img_processing_settings["target_imsize"]=target_imsize,
    img_processing_settings["target_channels"]=target_channels,
    img_processing_settings["normalization_method"]=normalization_method,
    img_processing_settings["mean_trainingdata"]=mean_trainingdata,
    img_processing_settings["std_trainingdata"]=std_trainingdata,
    img_processing_settings["zoom_factor"]=zoom_factor,
    img_processing_settings["zoom_interpol_method"]=zoom_interpol_method,
    img_processing_settings["padding_mode"]=padding_mode,
    
    return img_processing_settings



def forward_images_cv2(model_pb,img_processing_settings,images,pos_x,pos_y,pix):
    """
    Run inference on images using a tensorflow model
    Parameters
    ----------    
    model_pb: cv2.dnn_Net object; a frozen model graph that was loaded using cv2.dnn.readNet
    img_processing_settings: pd.DataFrame; a DataFrame generated using load_model_meta()
    images: numpy array of shape (nr.images,height,width) for grayscale images, or 
        of shape (nr.images,height,width,channels) for RGB images
    pos_x: float or ndarray of length N
        The x coordinate(s) of the centroid of the event(s) [um]
    pos_y: float or ndarray of length N
        The y coordinate(s) of the centroid of the event(s) [um]
    pix: float
        Resolution [µm/pix]
    Returns
    ----------    
    pd.DataFrame ; A DataFrame with the following keys:
        target_imsize: input image size required by the neural net
        target_channels: number of image channels required by the neural net
        normalization_method: the method to normalize the pixel-values of the images
        mean_trainingdata: the mean pixel value obtained from the training dataset
        std_trainingdata: the std of the pixel values obtained from the training dataset
        zoom_factor: factor by which the size of the images should be zoomed
        zoom_interpol_method: OpenCV interpolation flag
        padding_mode: OpenCV borderType flag
    """

    target_imsize = int(img_processing_settings["target_imsize"].values[0])
    target_channels = int(img_processing_settings["target_channels"].values[0])
    zoom_factor = float(img_processing_settings["zoom_factor"].values[0])
    zoom_interpol_method = str(img_processing_settings["zoom_interpol_method"].values[0])
    padding_mode = str(img_processing_settings["padding_mode"].values[0])
    normalization_method = str(img_processing_settings["normalization_method"].values[0])
    mean_trainingdata = img_processing_settings["mean_trainingdata"].values[0]
    std_trainingdata = img_processing_settings["std_trainingdata"].values[0]

    #Preprocess images
    images = image_preprocessing(images,pos_x=pos_x,pos_y=pos_y,pix=pix,
                                 target_imsize=target_imsize,
                                 target_channels=target_channels,
                                 zoom_factor=zoom_factor,
                                 zoom_interpol_method=zoom_interpol_method,
                                 padding_mode=padding_mode,
                                 normalization_method=normalization_method,
                                 mean_trainingdata=mean_trainingdata,
                                 std_trainingdata=std_trainingdata)

    #Load the model
    blob = cv2.dnn.blobFromImages(images, 1, (target_imsize,target_imsize), swapRB=False, crop=False)
    model_pb.setInput(blob)
    output_pb = model_pb.forward()
    return output_pb
    
    
    

def pad_arguments_np2cv(padding_mode):
    """
    NumPy's pad and OpenCVs copyMakeBorder can do the same thing, but the 
    function arguments are called differntly.
    This function takes numpy padding_mode argument and returns the 
    corresponsing borderType for cv2.copyMakeBorder
    Parameters
    ---------- 
    padding_mode: str; numpy padding mode
        - "constant" (default): Pads with a constant value.
        - "edge": Pads with the edge values of array.
        - "linear_ramp": Pads with the linear ramp between end_value and the array edge value.
        - "maximum": Pads with the maximum value of all or part of the vector along each axis.
        - "mean": Pads with the mean value of all or part of the vector along each axis.
        - "median": Pads with the median value of all or part of the vector along each axis.
        - "minimum": Pads with the minimum value of all or part of the vector along each axis.
        - "reflect": Pads with the reflection of the vector mirrored on the first and last values of the vector along each axis.
        - "symmetric": Pads with the reflection of the vector mirrored along the edge of the array.
        - "wrap": Pads with the wrap of the vector along the axis. The first values are used to pad the end and the end values are used to pad the beginning.
    Returns
    ----------   
    str: OpenCV borderType     
        - "cv2.BORDER_CONSTANT": iiiiii|abcdefgh|iiiiiii with some specified i 
        - "cv2.BORDER_REFLECT": fedcba|abcdefgh|hgfedcb
        - "cv2.BORDER_REFLECT_101": gfedcb|abcdefgh|gfedcba
        - "cv2.BORDER_DEFAULT": same as BORDER_REFLECT_101
        - "cv2.BORDER_REPLICATE": aaaaaa|abcdefgh|hhhhhhh
        - "cv2.BORDER_WRAP": cdefgh|abcdefgh|abcdefg
    """

    #Check that the padding_mode is actually supported by OpenCV
    supported = ["constant","edge","reflect","symmetric","wrap"]
    assert padding_mode in supported, "The padding mode: '"+padding_mode+"' is\
 not supported"
    
    if padding_mode=="constant":
        return "cv2.BORDER_CONSTANT"
    if padding_mode=="edge":
        return "cv2.BORDER_REPLICATE"
    if padding_mode=="reflect":
        return "cv2.BORDER_REFLECT_101"
    if padding_mode=="symmetric":
        return "cv2.BORDER_REFLECT"
    if padding_mode=="wrap":
        return "cv2.BORDER_WRAP"
    
    
def zoom_arguments_scipy2cv(zoom_factor,zoom_interpol_method):
    """
    Resulting images after performing ndimage.zoom or cv2.resize are never identical,
    but with certain settings you get at least similar results. 
    Parameters
    ----------    
    zoom_factor: float, 
        factor by which the size of the images should be zoomed
    zoom_interpol_method: int, 
        The order of the spline interpolation
    
    Returns
    ----------    
    str; OpenCV interpolation flag
    """
    opencv_zoom_options = ["cv2.INTER_NEAREST","cv2.INTER_LINEAR","cv2.INTER_AREA","cv2.INTER_CUBIC","cv2.INTER_LANCZOS4"]
    if type(zoom_interpol_method)==str:
        if zoom_interpol_method in opencv_zoom_options:
            return zoom_interpol_method
    
    if zoom_factor>=0.8:
        if zoom_interpol_method==0: return "cv2.INTER_NEAREST"
        elif zoom_interpol_method==1: return "cv2.INTER_LINEAR"
        elif zoom_interpol_method==2: return "cv2.INTER_CUBIC"
        elif zoom_interpol_method==3: return "cv2.INTER_LANCZOS4"
        elif zoom_interpol_method==4: return "cv2.INTER_LANCZOS4"
        elif zoom_interpol_method==5: return "cv2.INTER_LANCZOS4"

    if zoom_factor<0.8: #for downsampling the image, all methods perform similar
        #but cv2.INTER_LINEAR, is closest most of the time, irrespective of the zoom_order
        return "cv2.INTER_LINEAR"
    

    
